unit webcam;

interface

uses Windows, DirectShow9,ActiveX,dialogs,sysutils,classes,graphics,syncobjs,opengl,textures;

type TDevice_=record
              name:string;
              moniker:imoniker;
             end;

type TDevices_=array of TDevice_;

type TMode_=record
            w:integer;
            h:integer;
            bpp:integer;
            str:string;
           end;
type TModes_=array of TMode_;

type
  TCapturedBitmap = class(TBitmap); // to eliminate some problems with BCB


const IID_IPropertyBag: TGUID = '{55272A00-42CB-11CE-8135-00AA004BB851}';

type tsamplegrabbercb = class(tobject, isamplegrabbercb)
      private
       FBuffer : PByte;
       _CS:TCriticalSection;
       _grabbed:boolean;
      public
       function QueryInterface(const iid: tguid; out obj): hresult; stdcall;
       function _AddRef: integer; stdcall;
       function _Release: integer; stdcall;
       function SampleCB(sampletime: double; psample: imediasample): hresult; stdcall;
       function BufferCB(sampletime: double; pbuffer: pbyte; bufferlen: longint): hresult; stdcall;
      end;




type TWebCamera=class(TObject)
      private
       _CameraNames:TStringList;
       _ModeNames:TStringList;
       _CameraMoniker:imoniker;
       _mediaControl: imediacontrol;
       _cam_w:integer;
       _cam_h:integer;
       _now_camera:integer;
       _now_mode:integer;
      Builder: ICaptureGraphBuilder2;
      VCap:IBaseFilter;
      Graph: IGraphBuilder;
      grabber: isamplegrabber;
      VStreamConf: IAMStreamConfig;   // for video capture
      fgrabber: ibasefilter;
      mt: tammediatype;
      ProcAmp:IAMVideoProcAmp;
      CamControl:IAMCameraControl;
      hr:hresult;
      NullRender:ibasefilter;

      bright_min:integer;
      bright_max:integer;
      bright_dt:integer;
      bright_val:integer;

      exp_min:integer;
      exp_max:integer;
      exp_dt:integer;
      exp_val:integer;


      protected
       procedure GetCamerasList(new_camera:integer=0);
       procedure GetCameraModes(new_mode:integer=0);
       procedure StartGraph;
       procedure StopGraph;
      public
       grabbercb:TSampleGrabbercb;
       constructor Create; virtual;
       destructor Destroy; override;
       procedure Connect(new_camera:integer=0;new_mode:integer=0);
       procedure ChangeMode(new_mode:integer=0);
       function GetBitmap(capturedImage:TCapturedBitmap; tme:int64):boolean;
       function GetGLtexture:GLUINT;
       property CameraNames: TStringList read _CameraNames;
       property ModeNames:   TStringList read _ModeNames;

       procedure SetBrightness(new_bright:integer);
       procedure SetExposure(new_exp:integer);


end;








var aaa,bbb:integer;



var WebCamera:TWebCamera;
var ID:GLUINT=0;
var tmp:packed array [0..128*128*3-1] of byte;

implementation

procedure TWebCamera.SetBrightness(new_bright:integer);
var need:integer;
begin

need:=round((new_bright/100)*(bright_max-bright_min));

if procAmp<>nil then
ProcAmp.Set_(VideoProcAmp_Brightness, need, VideoProcAmp_Flags_Manual);

{
ProcAmp.Set_(VideoProcAmp_Brightness, 7500, VideoProcAmp_Flags_Manual);
hr:=CamControl.Set_(CameraControl_Exposure,0,CameraControl_Flags_Manual);
}

//if failed(hr) then showmessage('Error 10 '+inttohex(hr,8));
{if Succeeded(pProcAmp.Set_(VideoProcAmp_Brightness, 7500, VideoProcAmp_Flags_Manual)) then
  if Succeeded(pProcAmp.Set_(VideoProcAmp_Hue, 2500, VideoProcAmp_Flags_Manual)) then
  if Succeeded(pProcAmp.Set_(VideoProcAmp_Saturation, 10000, VideoProcAmp_Flags_Manual)) then}



end;


procedure TWebCamera.SetExposure(new_exp:integer);
var need:integer;
begin

need:=round((new_exp/100)*(exp_max-exp_min));
if camControl<>nil then
hr:=CamControl.Set_(CameraControl_Exposure,need,CameraControl_Flags_Manual);
showmessage(inttohex(hr,8));

end;

function TWebCamera.GetGLTexture:GLUINT;
var x,y,i,j:integer;
var tmpR,tmpG,tmpB:pByte;
begin

    try
     grabbercb._CS.Enter;

     if not grabbercb._grabbed then
      begin


      if ID=0 then
       begin
        glGenTextures(1, @ID);
        glBindTexture(GL_TEXTURE_2D, ID);
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        Result := ID;
       end;
//       fillchar(tmp,0,sizeof(tmp));

        for x := 0 to _cam_w - 1 do
         for y := 0 to _cam_h - 1 do
           if Integer(grabbercb.FBuffer)<>0 then
          begin

           tmpR:=pointer(Integer(grabbercb.FBuffer)+x*3+y*3*_cam_w);
           tmpG:=pointer(Integer(grabbercb.FBuffer)+x*3+y*3*_cam_w+1);
           tmpB:=pointer(Integer(grabbercb.FBuffer)+x*3+y*3*_cam_w+2);

           i:=round(127*x/_cam_w);
           j:=round(127*y/_cam_h);

           tmp[j*3*128+i*3]:=tmpR^;
           tmp[j*3*128+i*3+1]:=tmpG^;;
           tmp[j*3*128+i*3+2]:=tmpB^;;
          end;




        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, 128, 128, 0, GL_BGR, GL_UNSIGNED_BYTE, @tmp);





      grabbercb._grabbed:=true;
      end else result:=0
          finally
      grabbercb._CS.Leave;

      end;



end;

function TWebCamera.GetBitmap(capturedImage:TCapturedBitmap; tme:int64):boolean;
var  BitmapInfo: TBITMAPINFO;
  usage: cardinal;
begin

    try
     grabbercb._CS.Enter;

     if not grabbercb._grabbed then
      begin

   ZeroMemory(@BitmapInfo, sizeof(BitmapInfo));
   BitMapInfo.bmiHeader.biSize:=SizeOf(TBitmapInfo);
   BitMapInfo.bmiHeader.biWidth:=_cam_w;
   BitMapInfo.bmiHeader.biHeight:=_cam_h;
   BitMapInfo.bmiHeader.biBitCount:=24;
   BitMapInfo.bmiHeader.biPlanes:=1;
   BitMapInfo.bmiHeader.biCompression:=BI_RGB;
   BitMapInfo.bmiHeader.biSizeImage:=_cam_w*_cam_h*4;

   CapturedImage.PixelFormat:= pf32bit;
   CapturedImage.Width:= _cam_w;  // enlarge image to ensure the whole DIB gets copied
   CapturedImage.Height:= _cam_h;
   usage:= DIB_RGB_COLORS;  // BITMAP contains real color values



        CapturedImage.Canvas.Lock;
        result:= StretchDIBits( CapturedImage.Canvas.Handle, // handle of device context
                                     0, 0, _cam_w, _cam_h,                  // destination rectangle
                                     0, 0, _cam_w, _cam_h,                  // source rectangle
                                     grabbercb.FBuffer,      // address of bitmap bits
                                     BitmapInfo,	          // address of bitmap data
                                     usage,
                                     SRCCOPY ) <> 0;
        CapturedImage.Canvas.Unlock;
      result:=true;
      grabbercb._grabbed:=true;
      end else result:=false;
          finally
      grabbercb._CS.Leave;

      end;



end;

constructor TWebCamera.Create;
begin
 _CameraNames:=TStringList.Create;
 _ModeNames:=TStringList.Create;
 _now_camera:=-1;
 _now_mode:=-1;

  GrabberCB:=TSampleGrabberCB.Create;
  GrabberCB._CS:=TCriticalSection.Create;
end;

destructor TWebCamera.Destroy;
begin
  StopGraph;
  _CameraNames.Free;
  _ModeNames.Free;
  GrabberCB._CS.Enter;
  GrabberCB._CS.Leave;
  GrabberCB._CS.Free;
  GrabberCB.Free
end;






function checkguid(const p1, p2: tguid): boolean;
var
  i: byte;
begin
  result := false;
  for i := 0 to 7 do if p1.d4[i] <> p2.d4[i] then exit;
  result := (p1.d1 = p2.d1) and (p1.d2 = p2.d2) and (p1.d3 = p2.d3);
end;

procedure FreeMediaType(mt: TAMMEDIATYPE);
begin
  if mt.cbFormat <> 0 then begin
    CoTaskMemFree(mt.pbFormat);
    // Strictly unnecessary but tidier
    mt.cbFormat:= 0;
    mt.pbFormat:= nil;
  end;
  mt.pUnk:= nil;
end;


procedure DeleteMediaType(pmt: PAMMEDIATYPE);
begin
  // allow NULL pointers for coding simplicity
  if pmt <> nil then begin
    FreeMediaType(pmt^);
    CoTaskMemFree(pmt);
  end;
end;



procedure TWebCamera.GetCameraModes(new_mode:integer=0);
var
  capCount, capSize: integer;
  i:integer;
  pMediaType: PAMMEDIATYPE;
  vConfig: TVideoStreamConfigCaps; //?
  bih: TBitmapInfoHeader;
  fps: integer;
begin

 _modeNames.Clear;

 if (VStreamConf.GetNumberOfCapabilities(capCount, capSize)=S_OK) then
 if (new_mode<capCount) then
  for i := 0 to capCount - 1 do
   if VStreamConf.GetStreamCaps(i, pMediaType, vConfig)=S_OK then
     begin
        if CheckGUID(pMediaType^.formattype, FORMAT_VideoInfo) then
         begin
          bih:= PVideoInfoHeader(pmediatype^.pbFormat)^.bmiHeader;
          fps:=PVideoInfoHeader(pmediatype^.pbFormat)^.AvgTimePerFrame;
          if fps<>0 then fps:=round(10000000/fps);
         end
          else
         begin
          bih:= PVideoInfoHeader2(pmediatype^.pbFormat)^.bmiHeader;
          fps:=PVideoInfoHeader2(pmediatype^.pbFormat)^.AvgTimePerFrame;
          if fps<>0 then fps:=round(10000000/fps);
         end;

         if (i=new_mode) and not failed(VStreamConf.SetFormat(pMediaType^)) then
          begin
           _cam_w:=bih.biWidth;
           _cam_h:=bih.biHeight;
          end;

       _modeNames.Add(inttostr(bih.biWidth)+'x'+
                     inttostr(bih.biHeight)+'x'+
                     inttostr(bih.biBitCount)+' ('+
                     inttostr(fps)+'fps)');

         DeleteMediaType(pMediaType);
        end;
      

end;


procedure TWebCamera.StartGraph;
begin
 hr:=Graph.addfilter(vcap, 'SomeVideoCapture'); if failed(hr) then showmessage('Error 5');
 hr:=Graph.addfilter(fgrabber, 'SomeGrabber');if failed(hr) then showmessage('Error 10a');
 hr:=Graph.addfilter(nullrender, 'SomeNullRender');if failed(hr) then showmessage('Error 10b');
 hr:=Builder.RenderStream(nil, @MEDIATYPE_Video, vcap, fgrabber, nullrender);
 if failed(hr) then showmessage('Error 14 ');
 hr:=_mediaControl.Run;
end;

procedure TWebCamera.StopGraph;
var someFilter:IBaseFilter;
var filters:IEnumFilters;
var __mediacontrol:imediacontrol;
begin


_mediaControl.Stop;

grabbercb._cs.Enter;

graph.RemoveFilter(vcap);
graph.RemoveFilter(fgrabber);
graph.RemoveFilter(nullrender);

graph.EnumFilters(filters);

while filters.Next(1, someFilter,nil)=S_OK do
 begin
  graph.RemoveFilter(someFilter);
  filters.Reset;
 end;

grabbercb._cs.Leave;
end;


procedure TWebCamera.ChangeMode(new_mode:integer=0);
begin
 StopGraph;
 GetCameraModes(new_mode);
 StartGraph;
end;



procedure TWebCamera.Connect(new_camera:integer=0;new_mode:integer=0);
var VideoProcAmpFlags   : TVideoProcAmpFlags;
var CameraControlFlags  : TCameraControlFlags;

begin




   if vcap<>nil then StopGraph;




   GetCamerasList(new_camera);


   hr:=_CameraMoniker.BindToObject(nil, nil, iid_ibasefilter, vcap);
   if (VCap = nil) or failed(hr) then ShowMessage('Error 1');



    if failed(CoCreateInstance(CLSID_FilterGraph,
                     nil,
                     CLSCTX_INPROC,
                     IID_IGraphBuilder,
                     Graph)) then showmessage('Error 2');

    if failed(CoCreateInstance(CLSID_CaptureGraphBuilder2,
                     nil,
                     CLSCTX_INPROC,
                     IID_ICaptureGraphBuilder2,
                     Builder)) then showmessage('Error 3');

     if failed(Builder.SetFiltergraph(Graph)) then showmessage('Error 4');


    // !!! What if this interface isn't supported?
    // we use this interface to set the frame rate and get the capture size
    hr:= Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Interleaved,
                                VCap, IID_IAMStreamConfig, VStreamConf);
    if hr <> NOERROR then
    if failed(Builder.FindInterface( @PIN_CATEGORY_CAPTURE, @MEDIATYPE_Video,
                                  VCap, IID_IAMStreamConfig, VStreamConf)) then Showmessage('some-error');




    GetCameraModes(new_mode);




hr:= CoCreateInstance(CLSID_NullRenderer,
                      nil,
                      CLSCTX_INPROC,
                       IID_IBaseFilter,
                       NullRender);if failed(hr) then showmessage('Error 9');

hr:=CoCreateInstance(clsid_samplegrabber,
                     nil,
                     clsctx_inproc,
                     iid_ibasefilter,
                     fgrabber);if failed(hr) then showmessage('Error 9');

hr:=fgrabber.queryinterface(iid_isamplegrabber, grabber);if failed(hr) then showmessage('Error 11');



zeromemory(@mt, sizeof(mt));
mt.majortype := mediatype_video;
mt.subtype := MEDIASUBTYPE_RGB24; {MEDIASUBTYPE_YUY2};
mt.formattype := format_videoinfo;

hr:=grabber.setmediatype(mt);
if failed(hr) then showmessage('Error 12');

Grabber.SetBufferSamples(false);
Grabber.SetOneShot(false);
Grabber.SetCallback(TSampleGrabberCB(GrabberCB), 0);



if failed(vcap.QueryInterface(IID_IAMVideoProcAmp, ProcAmp)) then ProcAmp:=nil;
if failed(vcap.QueryInterface(IID_IAMCameraControl,CamControl)) then CamControl:=nil;
if failed(graph.queryinterface(iid_imediacontrol, _mediaControl)) then ShowMessage('error mediacontrol');


StartGraph;









if procAmp<>nil then
ProcAmp.GetRange(VideoProcAmp_Brightness,
                 bright_min,
                 bright_max,
                 bright_dt,
                 bright_val,
                 VideoProcAmpFlags);

if CamControl<>nil then
CamControl.GetRange(CameraControl_Exposure,
                 exp_min,
                 exp_max,
                 exp_dt,
                 exp_val,
                 CameraControlFlags);


end;


procedure TWebCamera.GetCamerasList(new_camera:integer=0);
var
  sysdevenum: icreatedevenum;
  enumcat: ienummoniker;
  moniker: imoniker;
  propbag: ipropertybag;
  varname: olevariant;
  i:integer;
begin
  CameraNames.clear;
  if (cocreateinstance(clsid_systemdeviceenum, nil, clsctx_inproc, iid_icreatedevenum, sysdevenum) = s_ok) then
    if (sysdevenum.createclassenumerator(clsid_videoinputdevicecategory{clsiddeviceclass}, enumcat, 0) = s_ok) then
      while enumcat.next(1, moniker, nil) = s_ok do
        if (moniker.bindtostorage(nil, nil, iid_ipropertybag, propbag) = s_ok) and (propbag <> nil) then
          begin
            propbag.read('FriendlyName', varname, nil);

            CameraNames.add(varname);
            if new_camera=CameraNames.Count-1 then
             _CameraMoniker:=moniker;
           end;
end;


function tsamplegrabbercb.samplecb(sampletime: double; psample: imediasample): hresult;
var sampleBuffer:pByte;
    BufferLen:integer;
begin
   pSample.GetPointer(sampleBuffer);
   BufferLen:=pSample.GetActualDataLength();
   aaa:=BufferLen;

  _CS.Enter;
  try
    ReallocMem(FBuffer,BufferLen);
    CopyMemory(FBuffer, sampleBuffer, BufferLen);
    _grabbed:=false;
  finally
  _CS.Leave;
  end;

aaa:=BufferLen;
  Result:= S_OK;

end;

function tsamplegrabbercb.buffercb(sampletime: double; pbuffer: pbyte; bufferlen: integer): hresult;
begin
  Result:= S_OK;
end;

function tsamplegrabbercb._addref: integer;
begin
  result := 2;
end;

function tsamplegrabbercb._release: integer;
begin
  result := 1;
end;


function tsamplegrabbercb.queryinterface(const iid: tguid; out obj): hresult;
begin
  // we need to return the two event interfaces when they're asked for
  result := e_nointerface;
  if checkguid(iid, isamplegrabbercb) or checkguid(iid, iunknown) then begin
    if getinterface(iid, obj) then result := s_ok;
  end;
end;



initialization



end.
